@@ -20,6 +20,7 @@ import (
 	"github.com/golang/glog"
 	"golang.org/x/time/rate"
 	corev1 "k8s.io/api/core/v1"
+	v1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/labels"
@@ -35,10 +36,10 @@ import (
 	"k8s.io/kubectl/pkg/drain"
 
 	// "k8s.io/client-go/kubernetes/scheme"
-	sriovnetworkv1 "github.com/openshift/sriov-network-operator/pkg/apis/sriovnetwork/v1"
-	snclientset "github.com/openshift/sriov-network-operator/pkg/client/clientset/versioned"
-	sninformer "github.com/openshift/sriov-network-operator/pkg/client/informers/externalversions"
-	"github.com/openshift/sriov-network-operator/pkg/utils"
+	sriovnetworkv1 "github.com/k8snetworkplumbingwg/sriov-network-operator/api/v1"
+	snclientset "github.com/k8snetworkplumbingwg/sriov-network-operator/pkg/client/clientset/versioned"
+	sninformer "github.com/k8snetworkplumbingwg/sriov-network-operator/pkg/client/informers/externalversions"
+	"github.com/k8snetworkplumbingwg/sriov-network-operator/pkg/utils"
 )
 
 const (
@@ -48,6 +49,8 @@ const (
 	// maxUpdateBackoff is the maximum time to react to a change as we back off
 	// in the face of errors.
 	maxUpdateBackoff = 60 * time.Second
+
+	UNSUPPORTED_NIC_ID_CONFIGMAP = "unsupported-nic-ids"
 )
 
 type Message struct {
@@ -60,6 +63,8 @@ type Daemon struct {
 	name      string
 	namespace string
 
+	platform utils.PlatformType
+
 	client snclientset.Interface
 	// kubeClient allows interaction with Kubernetes, including the node we are running on.
 	kubeClient *kubernetes.Clientset
@@ -88,6 +93,8 @@ type Daemon struct {
 
 	drainable bool
 
+	disableDrain bool
+
 	nodeLister listerv1.NodeLister
 
 	workqueue workqueue.RateLimitingInterface
@@ -126,9 +133,11 @@ func New(
 	stopCh <-chan struct{},
 	syncCh <-chan struct{},
 	refreshCh chan<- Message,
+	platformType utils.PlatformType,
 ) *Daemon {
 	return &Daemon{
 		name:       nodeName,
+		platform:   platformType,
 		client:     client,
 		kubeClient: kubeClient,
 		exitCh:     exitCh,
@@ -161,6 +170,63 @@ func New(
 	}
 }
 
+func (dn *Daemon) getUnsupportedNicIdConfigMap() (*v1.ConfigMap, error) {
+	cm, err := dn.kubeClient.CoreV1().ConfigMaps(namespace).Get(
+		context.Background(),
+		UNSUPPORTED_NIC_ID_CONFIGMAP,
+		metav1.GetOptions{},
+	)
+	return cm, err
+}
+
+func (dn *Daemon) annotateUnsupportedNicIdConfigMap(cm *v1.ConfigMap, nodeName string) (*v1.ConfigMap, error) {
+	jsonData, err := json.Marshal(cm.Data)
+	if err != nil {
+		return nil, err
+	}
+	cmData := string(jsonData)
+
+	annotationKey := "openshift.io/" + nodeName
+	annotationData, ok := cm.ObjectMeta.Annotations[annotationKey]
+	if ok && cmData == annotationData {
+		return cm, nil
+	}
+
+	if cm.ObjectMeta.Annotations == nil {
+		cm.ObjectMeta.Annotations = make(map[string]string)
+	}
+	cm.ObjectMeta.Annotations[annotationKey] = cmData
+	return dn.kubeClient.CoreV1().ConfigMaps(namespace).Update(context.Background(), cm, metav1.UpdateOptions{})
+}
+
+func (dn *Daemon) tryCreateUdevRuleWrapper() error {
+	var unsupportedNicIdMap map[string]string
+
+	// retrieve config map
+	cm, configMapErr := dn.getUnsupportedNicIdConfigMap()
+	if configMapErr != nil {
+		glog.V(2).Info("Could not retrieve Nic Id ConfigMap: ", configMapErr)
+	} else {
+		unsupportedNicIdMap = cm.Data
+	}
+
+	// update udev rule and if update succeeds
+	err := tryCreateUdevRule(unsupportedNicIdMap)
+	if err != nil {
+		return err
+	}
+
+	if configMapErr != nil {
+		return nil
+	}
+
+	if _, err := dn.annotateUnsupportedNicIdConfigMap(cm, dn.name); err != nil {
+		glog.V(2).Info("Could not update Nic Id ConfigMap", "err", err)
+	}
+
+	return nil
+}
+
 // Run the config daemon
 func (dn *Daemon) Run(stopCh <-chan struct{}, exitCh <-chan error) error {
 	glog.V(0).Info("Run(): start daemon")
@@ -169,9 +235,11 @@ func (dn *Daemon) Run(stopCh <-chan struct{}, exitCh <-chan error) error {
 	defer dn.workqueue.ShutDown()
 
 	tryEnableRdma()
-	if err := tryCreateUdevRule(); err != nil {
+
+	if err := dn.tryCreateUdevRuleWrapper(); err != nil {
 		return err
 	}
+
 	var timeout int64 = 5
 	dn.mu = &sync.Mutex{}
 	informerFactory := sninformer.NewFilteredSharedInformerFactory(dn.client,
@@ -240,6 +308,11 @@ func (dn *Daemon) Run(stopCh <-chan struct{}, exitCh <-chan error) error {
 				lastSyncError: err.Error(),
 			}
 			return err
+		case <-time.After(30 * time.Second):
+			glog.V(2).Info("Run(): period refresh")
+			if err := dn.tryCreateUdevRuleWrapper(); err != nil {
+				glog.V(2).Info("Could not create udev rule: ", err)
+			}
 		}
 	}
 }
@@ -346,6 +419,11 @@ func (dn *Daemon) operatorConfigChangeHandler(old, new interface{}) {
 		glog.Infof("Set log verbose level to: %d", level)
 		flag.Set("v", level.String())
 	}
+	newDisableDrain := newCfg.Spec.DisableDrain
+	if dn.disableDrain != newDisableDrain {
+		dn.disableDrain = newDisableDrain
+		glog.Infof("Set Disable Drain to: %t", dn.disableDrain)
+	}
 }
 
 func (dn *Daemon) nodeStateSyncHandler(generation int64) error {
@@ -388,32 +466,31 @@ func (dn *Daemon) nodeStateSyncHandler(generation int64) error {
 
 	reqReboot := false
 	reqDrain := false
+	mcoReboot := false
 	for k, p := range dn.LoadedPlugins {
 		d, r := false, false
 		if dn.nodeState.GetName() == "" {
 			d, r, err = p.OnNodeStateAdd(latestState)
-			glog.V(0).Infof("nodeStateSyncHandler(): plugin %s: reqDrain %v, reqReboot %v", k, d, r)
 		} else {
 			d, r, err = p.OnNodeStateChange(dn.nodeState, latestState)
-			glog.V(0).Infof("nodeStateSyncHandler(): plugin %s: reqDrain %v, reqReboot %v", k, d, r)
 		}
 		if err != nil {
 			glog.Errorf("nodeStateSyncHandler(): plugin %s error: %v", k, err)
 			return err
 		}
+
+		if p.Name() == "mco_plugin" && d && !r {
+			mcoReboot = true
+		}
+
+		glog.V(0).Infof("nodeStateSyncHandler(): plugin %s: reqDrain %v, reqReboot %v", k, d, r)
 		reqDrain = reqDrain || d
 		reqReboot = reqReboot || r
 	}
-	glog.V(0).Infof("nodeStateSyncHandler(): reqDrain %v, reqReboot %v", reqDrain, reqReboot)
+	glog.V(0).Infof("nodeStateSyncHandler(): reqDrain %v, reqReboot %v disableDrain %v", reqDrain, reqReboot, dn.disableDrain)
 
-	if reqDrain {
-		glog.Info("nodeStateSyncHandler(): drain node")
-		if err := dn.drainNode(dn.name); err != nil {
-			return err
-		}
-	}
 	for k, p := range dn.LoadedPlugins {
-		if k != GenericPlugin {
+		if k != GenericPlugin && k != McoPlugin {
 			err := p.Apply()
 			if err != nil {
 				glog.Errorf("nodeStateSyncHandler(): plugin %s fail to apply: %v", k, err)
@@ -421,6 +498,21 @@ func (dn *Daemon) nodeStateSyncHandler(generation int64) error {
 			}
 		}
 	}
+	if mcoReboot {
+		glog.V(0).Infof("nodeStateSyncHandler(): MCO will trigger reboot")
+		err = dn.LoadedPlugins[McoPlugin].Apply()
+		if err != nil {
+			glog.Errorf("nodeStateSyncHandler(): mco_plugin fail to apply: %v", err)
+			return err
+		}
+		return nil
+	}
+	if reqDrain && !dn.disableDrain {
+		glog.Info("nodeStateSyncHandler(): drain node")
+		if err := dn.drainNode(dn.name); err != nil {
+			return err
+		}
+	}
 
 	if len(dn.LoadedPlugins) > 1 && !reqReboot {
 		// Apply generic_plugin last
@@ -538,8 +630,18 @@ func (dn *Daemon) restartDevicePluginPod() error {
 }
 
 func (dn *Daemon) loadVendorPlugins(ns *sriovnetworkv1.SriovNetworkNodeState) error {
-	pl := registerPlugins(ns)
-	pl = append(pl, GenericPlugin)
+	var pl []string
+
+	if dn.platform == utils.VirtualOpenStack {
+		pl = append(pl, VirtualPlugin)
+	} else {
+		pl = registerPlugins(ns)
+		if utils.ClusterType == utils.ClusterTypeOpenshift {
+			pl = append(pl, McoPlugin)
+		}
+		pl = append(pl, GenericPlugin)
+	}
+
 	dn.LoadedPlugins = make(map[string]VendorPlugin)
 
 	for _, pn := range pl {
@@ -720,24 +822,33 @@ func tryEnableRdma() (bool, error) {
 	return false, err
 }
 
-func tryCreateUdevRule() error {
+func tryCreateUdevRule(unsupportedNicIdMap map[string]string) error {
 	glog.V(2).Infof("tryCreateUdevRule()")
-	filePath := "/host/etc/udev/rules.d/10-nm-unmanaged.rules"
-	_, err := os.Stat(filePath)
-	if err != nil {
-		if os.IsNotExist(err) {
-			glog.V(2).Infof("tryCreateUdevRule(): file not existed, create file")
-			_, err := os.Create(filePath)
-			if err != nil {
-				glog.Errorf("tryCreateUdevRule(): fail to create file: %v", err)
-				return err
-			}
-		} else {
-			return err
-		}
+	dirPath := "/host/etc/udev/rules.d/"
+	filePath := dirPath + "10-nm-unmanaged.rules"
+
+	new_content := fmt.Sprintf("ACTION==\"add|change|move\", ATTRS{device}==\"%s\", ENV{NM_UNMANAGED}=\"1\"\n", strings.Join(sriovnetworkv1.GetMergedVfIds(unsupportedNicIdMap), "|"))
+
+	old_content, err := ioutil.ReadFile(filePath)
+	// if old_content = new_content, don't do anything
+	if err == nil && new_content == string(old_content) {
+		return nil
 	}
-	content := fmt.Sprintf("ACTION==\"add|change\", ATTRS{device}==\"%s\", ENV{NM_UNMANAGED}=\"1\"\n", strings.Join(sriovnetworkv1.VfIds, "|"))
-	err = ioutil.WriteFile(filePath, []byte(content), 0666)
+
+	glog.V(2).Infof("Old udev content '%v' and new content '%v' differ. Writing to file %v.",
+		strings.TrimSuffix(string(old_content), "\n"),
+		strings.TrimSuffix(new_content, "\n"),
+		filePath)
+
+	err = os.MkdirAll(dirPath, os.ModePerm)
+	if err != nil && !os.IsExist(err) {
+		glog.Errorf("tryCreateUdevRule(): failed to create dir %s: %v", dirPath, err)
+		return err
+	}
+
+	// if the file does not exist or if old_content != new_content
+	// write to file and create it if it doesn't exist
+	err = ioutil.WriteFile(filePath, []byte(new_content), 0666)
 	if err != nil {
 		glog.Errorf("tryCreateUdevRule(): fail to write file: %v", err)
 		return err
